#include <stdio.h>
#include <assert.h>
#include "ufbx.h"
#include "example_math.h"

// Manually computes `ufbx_node.node_to_parent`.
// NOTE: Does not account for ufbx-specific adjust transforms.
Matrix4 get_transform(ufbx_node *node)
{
    ufbx_props *props = &node->props;

    // Fetch the transform properties.
    // In practice these could be for example manually evaluated from animation curves.
    int64_t rotation_order  = ufbx_find_int(props, "RotationOrder", 0);
    Vector3 lcl_translation = ufbx_find_vec3(props, "Lcl Translation", ufbx_zero_vec3);
    Vector3 lcl_scaling     = ufbx_find_vec3(props, "Lcl Scaling", ufbx_zero_vec3);
    Vector3 lcl_rotation    = ufbx_find_vec3(props, "Lcl Rotation", ufbx_zero_vec3);
    Vector3 rotation_pivot  = ufbx_find_vec3(props, "RotationPivot", ufbx_zero_vec3);
    Vector3 scaling_pivot   = ufbx_find_vec3(props, "ScalingPivot", ufbx_zero_vec3);
    Vector3 rotation_offset = ufbx_find_vec3(props, "RotationOffset", ufbx_zero_vec3);
    Vector3 scaling_offset  = ufbx_find_vec3(props, "ScalingOffset", ufbx_zero_vec3);
    Vector3 pre_rotation    = ufbx_find_vec3(props, "PreRotation", ufbx_zero_vec3);
    Vector3 post_rotation   = ufbx_find_vec3(props, "PostRotation", ufbx_zero_vec3);

    // Convert rotations from Euler to quaternions.
    // `enum EulerOrder` here matches `ufbx_rotation_order`.
    // Pre/post rotation always use XYZ rotation order.
    Quaternion lcl_quat = Quaternion_euler(lcl_rotation, (EulerOrder)rotation_order);
    Quaternion pre_quat = Quaternion_euler(pre_rotation, EulerOrder_XYZ);
    Quaternion post_quat = Quaternion_euler(post_rotation, EulerOrder_XYZ);

    Matrix4 m = Matrix4_identity;

    // Local scaling with pivots and offsets
    m = Matrix4_translate(-scaling_pivot) * m;
    m = Matrix4_scale_nonuniform(lcl_scaling) * m;
    m = Matrix4_translate(scaling_pivot) * m;
    m = Matrix4_translate(scaling_offset) * m;

    // Rotations with pivot, PostRotation is always inverted.
    m = Matrix4_translate(-rotation_pivot) * m;
    m = Matrix4_rotate(Quaternion_inverse(post_quat)) * m;
    m = Matrix4_rotate(lcl_quat) * m;
    m = Matrix4_rotate(pre_quat) * m;
    m = Matrix4_translate(rotation_pivot) * m;
    m = Matrix4_translate(rotation_offset) * m;

    // Finally, translation
    m = Matrix4_translate(lcl_translation) * m;

    return m;
}

static bool failed = false;

void check_column(const char *label, Vector3 col, Vector3 ref)
{
    float error = 0.0f;
    error += fabsf(col.x - ref.x);
    error += fabsf(col.y - ref.y);
    error += fabsf(col.z - ref.z);
    if (error >= 0.001f) {
        failed = true;
    }

    printf("  %s: (%+.2f, %+.2f, %+.2f), reference (%+.2f, %+.2f, %+.2f)\n",
        label, col.x, col.y, col.z, ref.x, ref.y, ref.z);
}

int main(int argc, char **argv)
{
    if (argc < 2) {
        fprintf(stderr, "Usage: ./example <filename.fbx>\n");
        return 1;
    }
    ufbx_scene *scene = ufbx_load_file(argv[1], NULL, NULL);
    assert(scene);

    const char *prop_names[] = {
        "Lcl Translation",
        "Lcl Scaling",
        "Lcl Rotation",
        "RotationPivot",
        "ScalingPivot",
        "RotationOffset",
        "ScalingOffset",
        "PreRotation",
        "PostRotation",
    };

    for (ufbx_node *node : scene->nodes) {
        if (node->is_root) continue;

        printf("\n%s:\n", node->name.data);

        for (const char *name : prop_names) {
            ufbx_vec3 v = ufbx_find_vec3(&node->props, name, ufbx_zero_vec3);
            if (v.x != 0.0f || v.y != 0.0f || v.z != 0.0f) {
                printf("  %s: (%.2f, %.2f, %.2f)\n", name, v.x, v.y, v.z);
            }
        }

        int64_t order = ufbx_find_int(&node->props, "RotationOrder", 0);
        if (order != 0) {
            printf("  RotationOrder: %d\n", (int)order);
        }

        Matrix4 mat = get_transform(node);

        Vector3 col_x = { mat.m00, mat.m10, mat.m20 };
        Vector3 col_y = { mat.m01, mat.m11, mat.m21 };
        Vector3 col_z = { mat.m02, mat.m12, mat.m22 };
        Vector3 col_w = { mat.m03, mat.m13, mat.m23 };

        Vector3 ref_x = node->node_to_parent.cols[0];
        Vector3 ref_y = node->node_to_parent.cols[1];
        Vector3 ref_z = node->node_to_parent.cols[2];
        Vector3 ref_w = node->node_to_parent.cols[3];

        check_column("X", col_x, ref_x);
        check_column("Y", col_y, ref_y);
        check_column("Z", col_z, ref_z);
        check_column("W", col_w, ref_w);
    }

    if (failed) {
        fprintf(stderr, "Error: get_transform() does not match ufbx\n");
        return 1;
    }

    return 0;
}
